/*************************************************************************
 * This file is part of Crosroads 1.38, a load balancer and fail over
 * utility for TCP. Copyright (c) Karel Kubat, distributed under GPL.
 * Visit http://crossroads.e-tunity.com for information.
 *************************************************************************/
#include "crossroads.h"

void tcpserve (int server_sock) {
    fd_set set;
    int backend_sock, new, size, i, pid, backend_selected;
    struct sockaddr_in clientname;
    static int wakeup_started = 0;

    /* Set up our signal handlers. */
    if (!wakeup_started++) {
	/* Interruption signals */
	for (i = 0; relevant_sigs[i]; i++)
	    signal (relevant_sigs[i], interrupt);
	/* Child termination is ignored. */
	signal (SIGCHLD, SIG_IGN);
	/* Set wakeup handler for the wakeup calls. */
	if (activeservice->rev_interval) {
	    if ( (pid = fork()) < 0 )
		error ("Fork failed: %s", strerror(errno));
	    else if (!pid) {
		set_program_title ("Service %s: wakeup", activeservice->name);
		wakeup_handler();
	    } else {
		msg ("Service %s: started wakeup handler at pid %d",
		     activeservice->name, pid);
		servicereport->rev_pid = pid;
	    }
	}
    }
    
    msg ("Service %s: awaiting activity on port %d (socket fd %d)",
	 activeservice->name, activeservice->port, server_sock);

    if (listen (server_sock, activeservice->backlog + 1) < 0)
	error ("Service %s: failed to listen to server_socket: %s",
	       activeservice->name, strerror(errno));
    
    /* We're a service now. Never return, never exit
     * (unless something is REALLY wrong).
     */
    while (1) {
	/* Reload allow/deny files if requested so due to SIGHUP. */
	if (reload_allow_deny) {
	    reload_allow_deny = 0;
	    if (ipf_loadfile (activeservice->allowfile,
			      &(activeservice->allowchain),
			      &(activeservice->nallowchain)))
		warning ("Bad syntax in allow file");
	    if (ipf_loadfile (activeservice->denyfile,
			      &(activeservice->denychain),
			      &(activeservice->ndenychain)))
		warning ("Bad syntax in deny file");    
	    msg ("Service %s: allow chain has %d entries, "
		 "deny chain has %d entries", activeservice->name,
		 activeservice->nallowchain, activeservice->ndenychain);
	}
	    
	/* Block until we get a connection. */
	FD_ZERO (&set);
	FD_SET (server_sock, &set);
	if (select (FD_SETSIZE, &set, 0, 0, 0) < 0) {
	    msg ("Service %s: interrupt while waiting for activity: %d (%s)",
		 activeservice->name, errno, strerror(errno));
	    continue;
	}

	/* Accept the client-side connection. */
	size = sizeof(clientname);
	if ( (new = accept (server_sock, (struct sockaddr *) &clientname,
			    (socklen_t *) &size)) < 0 )
	    error ("Service %s: failure while accepting on server socket: %s",
		   activeservice->name, strerror(errno));

	/* Store client IP, it's used for lots of logging. */
	client_ip = inet_ntoa (clientname.sin_addr);
	msg ("Service %s: connection from %s, socket %d", 
	     activeservice->name, client_ip, new);
	if (ipf_denied ()) {
	    warning ("Service %s: %s matches deny list, "
		     "terminating connection", activeservice->name, client_ip);
	    close (new);
	    continue;
	}
	if (!ipf_allowed ()) {
	    warning ("Service %s: %s failes to match allow list, "
		     "terminating connection", activeservice->name, client_ip);
	    close (new);
	    continue;
	}

	/* Backend yet to be defined. */
	current_backend = -1;

	/* Incase of a http type service: handle separately. */
	if (activeservice->type == type_http) {
	    http_serve (new);
	    close (new);
	    continue;
	}
	
	/* This is an 'any' service type.
	 * Leave it alone if there are no back ends or if we exceed
	 * the max allowed clients. IN THAT CASE WE RETURN so that
	 * runservice() may close the listener socket, sleep, and
	 * create a new one.
	 */
	if (! backend_count()) {
	    warning ("Service %s: no back ends available",
		     activeservice->name);
	    close (new);
	    return;
	}

	/* Retry back ends until we succeed. */
	backend_selected = 0;
	while (!backend_selected) {
	    msg ("Service %s: About to choose a back end",
		 activeservice->name);
	    choose_backend();
	    if (current_backend < 0) {
		/* No back ends available now. NOTE: we return so that
		 * runservice() will retry in some time. */
		close (new);
		return;
	    }
	    
	    /* Connect to the backend. If this fails then we'll sleep
	     * and re-enter the backend selection loop instead of returning.
	     * In the loop we may need to decide to return after all. */
	    msg ("Service %s: trying back end %s, port %d",
		 activeservice->name, 
		 activeservice->backend[current_backend].server,
		 activeservice->backend[current_backend].port);
	    if ( (backend_sock = backend_connect()) < 0 ) {
		warning ("Service %s: failed to connect to server %s:%d "
			 "(ret %d)", activeservice->name,
			 activeservice->backend[current_backend].server,
			 activeservice->backend[current_backend].port,
			 backend_sock);
		continue;
	    }

	    /* Got a live one! */
	    backend_selected++;
	    if (fork_tcp_servicer (current_backend)) {
		/* We're the parent branch here. Close sockets so that
		 * we don't run out of file descriptors, and loop into
		 * the next select/accept run. */
		close (new);
		close (backend_sock);
	    } else {
		/* Child branch: piggyback to and fro.
		 * This one never returns. */
		incr_client_count();
		log_activity_start ();
		flag_verbose =
		    activeservice->backend[current_backend].verbosity;
		copysockets (new, backend_sock);
	    }
	}
    }
}
